import os
import subprocess
from typing import Union


class FijiRunner:  # pylint: disable=too-few-public-methods
    """Runs icy in headless with specified protocol and arguments

    Attributes:
        fiji_path (str | os.PathLike): Path to the fiji executable
            The executable can be found inside the installation folder of Fiji.
            Linux: Fiji.app/ImageJ-<os>.exe
            Windows: Fiji.app/ImageJ-<os>.exe
            MacOs: Fiji.app/Contents/MacOs/ImageJ-<os>

    """

    cmd = '{fiji} --ij2 --headless --console --run {script} "{args}"'

    def __init__(self, fiji_path: Union[str, os.PathLike]) -> None:
        self.fiji_path = fiji_path

        assert os.path.isfile(fiji_path), "Unable to found the given path"

    def run(self, script: Union[str, os.PathLike], **kwargs) -> int:
        """Runs fiji with the given script and additional kwargs

        Args:
            script (str | os.PathLike): Path to a Fiji script file
            **kwargs: Additional arguments given as key=value to the cmd line

        Returns:
            int: Return code of fiji

        """
        args = ",".join(f"{key}='{value}'" for key, value in kwargs.items())

        cmd = self.cmd.format(fiji=self.fiji_path, script=script, args=args)

        print("Calling Fiji with:", cmd)
        return subprocess.run(cmd, check=True, shell=True).returncode



"""Utilities for inputs/outputs with icy"""


import os
from typing import Collection, Union
import warnings
from xml.etree import ElementTree as ET
import zlib

import torch

import byotrack


def save_detections(detections_sequence: Collection[byotrack.Detections], path: Union[str, os.PathLike]) -> None:
    """Save a sequence of detections as valid rois for icy

    Format example:

    .. code-block:: xml

        <root>
            <roi>
                <classname>plugins.kernel.roi.roi2d.ROI2DArea</classname>
                <id>30</id>
                <name>spot #0</name>
                <selected>false</selected>
                <readOnly>false</readOnly>
                <properties>None</properties>
                <color>-16711936</color>
                <stroke>2</stroke>
                <opacity>0.3</opacity>
                <showName>false</showName>
                <z>0</z>
                <t>0</t>
                <c>-1</c>
                <boundsX>238</boundsX>
                <boundsY>486</boundsY>
                <boundsW>1</boundsW>
                <boundsH>2</boundsH>
                <boolMaskData>78:5e:63:64:4:0:0:5:0:3</boolMaskData>
            </roi>
            ...
        </root>

    Only needed tags are filled in the current implementation

    Args:
        detections_sequence (Collection[Detections]): Detections for each frame
        path (str | os.PathLike): Output path

    """
    ## XXX: Support for 3D images
    root = ET.Element("root")
    for detections in detections_sequence:
        for label, bbox in enumerate(detections.bbox):
            roi = ET.SubElement(root, "roi")
            i, j, height, width = bbox.tolist()
            mask = detections.segmentation[i : i + height, j : j + width] == label + 1

            ET.SubElement(roi, "classname").text = "plugins.kernel.roi.roi2d.ROI2DArea"
            ET.SubElement(roi, "t").text = str(detections.frame_id)
            ET.SubElement(roi, "z").text = "0"
            ET.SubElement(roi, "boundsX").text = str(j)
            ET.SubElement(roi, "boundsY").text = str(i)
            ET.SubElement(roi, "boundsW").text = str(width)
            ET.SubElement(roi, "boundsH").text = str(height)

            # The mask is converted into bytes and zipped
            compressed_bytes = zlib.compress(bytes(mask.reshape(height * width).numpy()), 2)

            # and then converted to the good string format: byte:byte:...:byte
            ET.SubElement(roi, "boolMaskData").text = ":".join(map(lambda byte: hex(byte)[2:], compressed_bytes))

    ET.ElementTree(root).write(path)


def load_tracks(path: Union[str, os.PathLike]) -> Collection[byotrack.Track]:
    """Load tracks in Icy format

    Format example:

    .. code-block:: xml

        <root>
            <trackfile version="1">
                <trackgroup description="mhtTracks-Run1">
                    <track id="-1743400864">
                        <detection classname="plugins.nchenouard.particletracking.DetectionSpotTrack" color="-6553856"
                            t="0" type="1" x="338.14285714285717" y="207.71428571428572" z="0"/>
                        <detection classname="plugins.nchenouard.particletracking.DetectionSpotTrack" color="-6553856"
                            t="1" type="1" x="338.5" y="207.5" z="0"/>
                        ...
                    </track>
                    ...
                </trackgroup>
            </trackfile>
        </root>

    For each point in each track the frame (time) and position (x, y, z) are given. An additional type precise if
    the detection is real or extrapolated. (Unused in ByoTrack)

    Args:
        path (str | os.PathLike): Input path

    Returns:
        Collection[Track]: Parsed tracks

    """
    tree = ET.parse(path)

    track_group = tree.find("trackgroup")
    assert track_group, "Track group not found in file"

    tracks = []

    for track in track_group:
        identifier = None
        if track.attrib.get("id"):
            identifier = int(track.attrib["id"])

        points = []
        frames = []
        unused_z = True
        for point in track:
            frames.append(int(point.attrib["t"]))
            points.append((float(point.attrib["y"]), float(point.attrib["x"]), float(point.attrib["z"])))
            if float(point.attrib["z"]) > 0:
                unused_z = False

        # Usually frames are sorted and consecutives. But we rather expect less to be more robust
        start = min(frames)
        end = max(frames) + 1

        points_tensor = torch.full((end - start, 3 - unused_z), torch.nan, dtype=torch.float32)
        points_tensor[torch.tensor(frames) - start] = torch.tensor(points)[:, : 3 - unused_z]

        tracks.append(byotrack.Track(start, points_tensor, identifier))

    return tracks


def save_tracks(tracks: Collection[byotrack.Track], path: Union[str, os.PathLike]) -> None:
    """Save tracks in Icy format

    .. warning:: Icy do not support partial tracks (track with undefined positions). Before calling this function
        you should first interpolate any missing position (See `ForwardBackwardInterpolater`)

    Format example:

    .. code-block:: xml

        <root>
            <trackfile version="1"/>
            <trackgroup description="mhtTracks-Run1">
                <track id="-1743400864">
                    <detection classname="plugins.nchenouard.particletracking.DetectionSpotTrack" color="-6553856"
                        t="0" type="1" x="338.14285714285717" y="207.71428571428572" z="0"/>
                    <detection classname="plugins.nchenouard.particletracking.DetectionSpotTrack" color="-6553856"
                        t="1" type="1" x="338.5" y="207.5" z="0"/>
                    ...
                </track>
                ...
            </trackgroup>
        </root>

    For each point in each track the frame (t) and position (x, y, z) are given. An additional type precise if
    the detection is real or extrapolated. (Unused in ByoTrack).

    Only needed tags are filled in the current implementation. (t, x, y, z)

    Args:
        path (str | os.PathLike): Input path

    Returns:
        Collection[Track]: Parsed tracks

    """
    root = ET.Element("root")
    ET.SubElement(root, "trackfile", {"version": "1"})
    track_group = ET.SubElement(root, "trackgroup", {"description": "ByoTrack"})
    for track in tracks:
        track_element = ET.SubElement(track_group, "track", {"id": str(track.identifier)})

        frame = track.start
        for point in track.points:
            if torch.isnan(point).any():
                warnings.warn(
                    "Found NaN in track points and Icy do not support partial tracks. "
                    "You should consider first using an interpolater on the tracks."
                )
                frame += 1
                continue  # No detection for this frame

            x = str(point[1].item())
            y = str(point[0].item())
            z = "-1"
            if point.shape[0] > 2:
                z = str(point[2].item())

            ET.SubElement(
                track_element,
                "detection",
                {
                    "classname": "plugins.nchenouard.particletracking.DetectionSpotTrack",
                    "t": str(frame),
                    "x": x,
                    "y": y,
                    "z": z,
                },
            )

            frame += 1

    ET.ElementTree(root).write(path)



import os
import pathlib
from typing import List, Union
import warnings

import numpy as np
from stardist.models import StarDist2D  # type: ignore
import torch

import byotrack
from byotrack.api.detector.detections import relabel_consecutive
from byotrack.api.parameters import ParameterBound


class StarDistDetector(byotrack.BatchDetector):
    """Runs stardist as a detector.

    Wraps the official implementation at https://github.com/stardist/stardist, following the paper:
    Uwe Schmidt, Martin Weigert, Coleman Broaddus, and Gene Myers. Cell Detection with Star-convex Polygons.
    International Conference on Medical Image Computing and Computer-Assisted Intervention (MICCAI),
    Granada, Spain, September 2018.

    We do not provide any code to train the stardist model. We expect you to already have trained the model.

    Note:
        This module requires `stardist` lib to be installed (with tensorflow). Please follow the instruction of the
        official implementation to install it.
    """

    progress_bar_description = "Detections (StarDist)"

    parameters = {
        "prob_threshold": ParameterBound(0.0, 1.0),
        "nms_threshold": ParameterBound(0.0, 1.0),
    }

    def __init__(self, model_dir: Union[str, os.PathLike], **kwargs) -> None:
        super().__init__(**kwargs)

        if self.batch_size != 1:  # TODO: Stardist do not give a straightforward implem for batch size > 1
            warnings.warn(
                "Current implementation do not support batch size greater than 1 and will iterate image by image"
            )

        path = pathlib.Path(model_dir)
        self._model = StarDist2D(None, path.name, str(path.parent))

        # Prop setter in stardist is changing type thus linting is failing...
        self.prob_threshold: float = self._model.thresholds.prob  # pylint: disable=no-member
        self.nms_threshold: float = self._model.thresholds.nms  # pylint: disable=no-member

    def detect(self, batch: np.ndarray) -> List[byotrack.Detections]:
        detections_list = []

        for i, image in enumerate(batch):
            segmentation, data = self._model.predict_instances(
                image, prob_thresh=self.prob_threshold, nms_thresh=self.nms_threshold, predict_kwargs={"verbose": 0}
            )

            detections_list.append(
                byotrack.Detections(
                    {
                        "segmentation": relabel_consecutive(torch.tensor(segmentation, dtype=torch.int32)),
                        "confidence": torch.tensor(data["prob"], dtype=torch.float32),
                        # Could use points data for position (but has been rounded to int, let's be more precise)
                        # "position": torch.tensor(data["points"], dtype=torch.float32),
                    },
                    frame_id=i,
                )
            )

        return detections_list



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Collection, Iterable

import numpy as np

import byotrack  # pylint: disable=cyclic-import
from .parameters import ParametrizedObjectMixin


class Tracker(ABC, ParametrizedObjectMixin):  # pylint: disable=too-few-public-methods
    """Base abstract tracker class

    A tracker can be run on a whole video and produces tracks

    """

    @abstractmethod
    def run(self, video: Iterable[np.ndarray]) -> Collection[byotrack.Track]:
        """Run the tracker on a whole video

        Args:
            video (Iterable[np.ndarray]): Sequence of frames (video)
                Each array is expected to have a shape (H, W, C)

        Returns:
            Collection[byotrack.Track]: Tracks of particles

        """


class MultiStepTracker(Tracker):  # pylint: disable=too-few-public-methods
    """Multi step tracker: split the tracking task into Detect / Link / Refine

    Attributes:
        detector (byotrack.Detector): Performs the detection on the video
        linker (byotrack.Linker): Links detections through time
        refiners (Iterable[byotrack.Refiner]): Optional refinement steps
            Empty by default

    """

    def __init__(
        self, detector: byotrack.Detector, linker: byotrack.Linker, refiners: Iterable[byotrack.Refiner] = ()
    ) -> None:
        super().__init__()

        self.detector = detector
        self.linker = linker
        self.refiners = refiners

    def run(self, video: Iterable[np.ndarray]) -> Collection[byotrack.Track]:
        detections = self.detector.run(video)
        tracks = self.linker.run(video, detections)
        for refiner in self.refiners:
            tracks = refiner.run(video, tracks)

        return tracks



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Collection, Iterable

import numpy as np

import byotrack  # pylint: disable=cyclic-import
from .parameters import ParametrizedObjectMixin


class Linker(ABC, ParametrizedObjectMixin):  # pylint: disable=too-few-public-methods
    """Base class for linkers in videos

    Link detections through time to build tracks.

    Each linker can define a set of parameters (See `ParametrizedObjectMixin`)

    """

    @abstractmethod
    def run(
        self, video: Iterable[np.ndarray], detections_sequence: Collection[byotrack.Detections]
    ) -> Collection[byotrack.Track]:
        """Run the linker on a whole video

        Args:
            video (Iterable[np.ndarray]): Sequence of frames (video)
                Each array is expected to have a shape (H, W, C)
            detections_sequence (Collection[byotrack.Detections]): Detections for each frame

        Returns:
            Collection[byotrack.Track]: Tracks of particles

        """



"""ByoTrack

Unified python API for biological particle tracking.

We provide a unified python API for tracking that can be easily
extended with new (and old) algorithms. We also provide
implementations of well-known algorithms following our API.

Overview:
* Video
    * Able to read classical format (supported by opencv) + tiff
* Particle Tracking
    * MultiStepTracker (Detect / Link / Refine)
* Particle Detections
    * Wavelet Detector [2] (Similar as the one in Icy [1] but coded in pytorch)
    * Stardist [3] (Inference only. Training should be done with the official implementation)
* Particle Linking
    * EMHT [4] (Wraps the implementation in Icy [1], requires Icy to be installed)
    * u-track / TrackMate [7] (Wraps TrackMate [6, 8] implementation in ImageJ/Fiji, requires Fiji to be installed)
* Tracks Refining
    * Cleaning
    * EMC2 [5]: Track stitching (gap closing)
    * Interpolate missing positions



Getting started:
----------------

```python
import byotrack

# Load some specific implementations
from byotrack.implementation.detector.wavelet import WaveletDetector
from byotrack.implementation.linker.icy_emht import IcyEMHTLinker
from byotrack.implementation.refiner.cleaner import Cleaner
from byotrack.implementation.refiner.stitching import EMC2Stitcher

# Read a video from a path, normalize and aggregate channels
video = byotrack.Video(video_path)
transform_config = VideoTransformConfig(aggregate=True, normalize=True, q_min=0.01, q_max=0.999)
video.set_transform(transform_config)

# Create a multi step tracker
## First the detector
## Smaller scale <=> search for smaller spots
## The noise threshold is linear with k. If you increase it, you will retrieve less spots.
detector = WaveletDetector(scale=1, k=3.0, min_area=5)

## Second the linker
## Hyperparameters are automatically chosen by Icy
linker = IcyEMHTLinker(icy_path)

## Finally refiners
## If needed you can add Cleaning and Stitching operations
refiners = []
if True:
    refiners.append(Cleaner(5, 3.5))  # Split tracks on position jumps and drop small ones
    refiners.append(EMC2Stitcher())  # Merge tracks if they track the same particle

tracker = byotrack.MultiStepTracker(detector, linker, refiners)

# Run the tracker
tracks = tracker.run(video)

# Save tracks
byotrack.Track.save(tracks, output_path)
```

Please refer to the ![official documentation](https://byotrack.readthedocs.io/en/latest/?badge=latest)

"""

# Import API and main features but no implementation
from byotrack.api.detector.detections import Detections
from byotrack.api.detector.detector import BatchDetector, Detector
from byotrack.api.linker import Linker
from byotrack.api.refiner import Refiner
from byotrack.api.tracker import MultiStepTracker, Tracker
from byotrack.api.tracks import Track
from byotrack.video import Video, VideoTransformConfig


__version__ = "0.4.1"



import dataclasses
import enum
import os
import pathlib
from typing import Collection, Iterable, Optional, Union
import warnings
from xml.etree import ElementTree as ET

import numpy as np

import byotrack
from byotrack import icy
from byotrack.api.parameters import ParameterEnum


PROTOCOL = pathlib.Path(__file__).parent / "emht_protocol.xml"
PARAMETRIZED_PROTOCOL = pathlib.Path(__file__).parent / "emht_protocol_with_full_specs.xml"


class Motion(enum.Enum):
    """Different motion models:

    Brownian: Random gaussian displacement at each time
    Directed: Random gaussian noise around a directed trajectory
    Multi: Switch randomly between Brownian and Directed

    """

    BROWNIAN = (0, 0)
    DIRECTED = (1, 0)
    MULTI = (0, 1)


@dataclasses.dataclass
class EMHTParameters:  # pylint: disable=too-many-instance-attributes
    """Parameters of EMHT algorithm [4]

    Some parameters are still not very clear to us, but we describe them to the best of our knowledge.

    Attributes:
        detections_fpr (float): Estimation of the false positive rate of the detection process
            Default: 0.1
        detections_fnr (float): Estimation of the false negative rate of the detection process
            Default: 0.1
        expected_track_length (int): Expected length of tracks. If negative, it defaults to the sequence size
            Default: -1
        expected_initial_particles (int): Estimation of the number of particles in the first frames. If negative, it
            defaults to the average number of detections by frame in the sequence
            Default: -1
        expected_new_particles (int): Estimation of the number of new particle by frame
            Default: 0
        existence_prob (float): Minimum probability to confirm a track existence
            Default: 0.5
        termination_prob (float): Minimum probability before terminating a track
            Default: 1e-4
        motion (Motion): Motion of the particles (Brownian vs Directed vs Multi). If MULTI, motion 1 is
            brownian and motion 2 is directed.
            Default: Motion.BROWNIAN
        use_most_likely_model (bool): Use the most likelymodel to predict rather than a weighted predictions
            Default: True
        update_motion_{1, 2} (bool): Update the covariance (Q ?) online of motion 1 or 2
        xy_std_{1,2} (float): Used to define the covariance (Q) of the process. (Looking at Icy code,
            it is not clear what they truly are, depending on the motion the covariance Q is set a bit weirdly)
            Default: 3.0
        z_std_{1,2} (float): Same as `xy_std` but for the 3d axes (if any)
            Default: 3.0
        inertia (float): Probability to not switch of motion model (For MULTI motion)
            Default: 0.8
        gate_factor (float): Max mahalanobis distance for potential association
            Default: 4.0
        tree_depth (int): Number of frames to consider in the search tree
            Default: 4
    """

    # detections
    detecions_fpr: float = 0.1
    detecions_fnr: float = 0.1

    # Partcles life
    expected_track_length: int = -1  # -1 => Default to size of the sequence
    expected_initial_particles: int = -1  # -1 => Default to average num of detections
    expected_new_particles: int = 10
    existence_prob: float = 0.5
    termination_prob: float = 1e-4

    # Motion
    motion: Motion = Motion.BROWNIAN
    use_most_likely_model: bool = True
    update_motion_1: bool = False  # Update std online for motion 1?
    update_motion_2: bool = False  # Update std online for motion 2?
    xy_std_1: float = 3.0  # Correspond to the first motion (brownian if multiple)
    z_std_1: float = 3.0  # Correspond to the first motion (brownian if multiple)
    xy_std_2: float = 3.0  # Correspond to the second motion (Always directed)
    z_std_2: float = 3.0  # Correspond to the second motion (Always directed)
    inertia: float = 0.8  # Prob to keep the same motion model

    # MHT
    gate_factor: float = 4.0
    tree_depth: int = 4

    def to_xml(self, detections_sequence: Collection[byotrack.Detections]) -> ET.ElementTree:
        """Convert to xml format used by Icy

        Format example:

        .. code-block:: xml

            <root>
                <MHTconfiguration>
                    <detectionInput detectionRate="0.9" detectionSource="noSource" numberFalseDetection="20"/>
                    <targetExistence confirmationThreshold="0.5" terminationThreshold="1.0E-4" trackLength="200"/>
                    <motionModel IMMinertia="0.8" isDirectedMotion_1="true" isDirectedMotion_2="true"
                        isMostLikelyModel="true" isSingleMotion="false" updateCovariance_2="false" updateMotion_1="true"
                        yxDisplacement_1="2" yxDisplacement_2="3" zDisplacement_1="2" zDisplacement_2="3"/>
                    <mht gateFactor="4" mhtDepth="4" numberNewObjects="10" numberObjectsFirstFrame="50"/>
                    <ouput trackGroupName="mht-tracks-1"/>
                </MHTconfiguration>
            </root>

        Args:
            detections_sequence (Collection[byotrack.Detections]): Detections (Some parameters requires to
                know the number of detections)

        Returns:
            xml.etree.ElementTree.ElementTree: Xml tree of the configuration

        """
        num_frames = (
            max(detections_sequence, key=lambda detections: detections.frame_id).frame_id
            - min(detections_sequence, key=lambda detections: detections.frame_id).frame_id
        )
        mean_particles = sum(map(lambda detections: detections.length, detections_sequence)) / len(detections_sequence)

        root = ET.Element("root")
        config = ET.SubElement(root, "MHTconfiguration")
        ET.SubElement(
            config,
            "detectionInput",
            {
                "detection_rate": str(1 - self.detecions_fnr),
                "detectionSource": "noSource",
                "numberFalseDetection": str(int(mean_particles * self.detecions_fpr)),
            },
        )

        ET.SubElement(
            config,
            "targetExistence",
            {
                "confirmationThreshold": str(self.existence_prob),
                "terminationThreshold": str(self.termination_prob),
                "trackLength": str(num_frames if self.expected_track_length < 0 else self.expected_track_length),
            },
        )

        ET.SubElement(
            config,
            "motionModel",
            {
                "isSingleMotion": "false" if self.motion is Motion.MULTI else "true",
                "isMostLikelyModel": "true" if self.use_most_likely_model else "false",
                "updateMotion_1": "true" if self.update_motion_1 else "false",
                "updateCovariance_2": "true" if self.update_motion_2 else "false",
                "isDirectedMotion_1": "true" if self.motion is Motion.DIRECTED else "false",
                "isDirectedMotion_2": "true" if self.motion is Motion.MULTI else "false",
                "IMMinertia": str(self.inertia),
                "yxDisplacement_1": str(self.xy_std_1),
                "yxDisplacement_2": str(self.xy_std_2),
                "zDisplacement_1": str(self.z_std_1),
                "zDisplacement_2": str(self.z_std_2),
            },
        )

        ET.SubElement(
            config,
            "mht",
            {
                "gateFactor": str(self.gate_factor),
                "mhtDepth": str(self.tree_depth),
                "numberNewObjects": str(self.expected_new_particles),
                "numberObjectsFirstFrame": str(
                    mean_particles if self.expected_initial_particles < 0 else self.expected_initial_particles
                ),
            },
        )

        return ET.ElementTree(root)


class IcyEMHTLinker(byotrack.Linker):  # pylint: disable=too-few-public-methods
    """Run EMHT [4] from Icy [1]

    This code is only a wrapper arounds Icy implementation as EMHT is painful to implement.

    About EMHT:
    It is a probabilistic tracking that uses statistical motion model on particles. It uses multiple
    kalman filters for each particle allowing a particle to have several mode of motions. It also keeps
    multiple hypothesis of tracking at each frame so that a final detections linking decision is made
    after seeing several frames in the past and future of these detections.

    Here we rely on the handmade protocol "emht_protocol" that expects detections as a valid rois file
    for icy and some hyperparameters.

    The workflow is:

    1. Detections to rois in Icy format
    2. Run the Icy protocol
    3. [In ICY] Read rois, estimate or load emht parameters, run emht, export tracks to xml
    4. Read Icy tracks and return

    Note:
        This implementation requires Icy to be installed (https://icy.bioimageanalysis.org/download/)

    Attributes:
        runner (byotrack.icy.IcyRunner): Icy runner
        motion (Motion): Prior on the underlying motion model (Brownian vs Directed vs Both)
            Given to the Icy block that estimates EMHT parameters. (See `full_specs` and `EMHTParameters` to have
            a finegrained control over the algorithm parameters)
            Default: Motion.BROWNIAN
        full_specs (Optional[EMHTParameters]): Full specification of the algorithm. If not provided,
            we use the estimation of EMHT parameters provided by Icy (with `motion` the only parameter to set).

    """

    parameters = {"motion": ParameterEnum({Motion.BROWNIAN, Motion.DIRECTED, Motion.MULTI})}
    rois_file = "_tmp_rois.xml"
    parameters_file = "_tmp_parameters.xml"
    tracks_file = "_tmp_tracks.xml"

    def __init__(
        self, icy_path: Optional[Union[str, os.PathLike]] = None, full_specs: Optional[EMHTParameters] = None
    ) -> None:
        """Constructor

        Args:
            icy_path (str | os.PathLike): Path to the icy jar (Icy is called with java -jar <icy_jar>)
                If not given, icy is searched in the PATH
        """
        super().__init__()
        self.runner = icy.IcyRunner(icy_path)
        self.motion = full_specs.motion if full_specs else Motion.BROWNIAN
        self.full_specs = full_specs

    def run(
        self, video: Iterable[np.ndarray], detections_sequence: Collection[byotrack.Detections]
    ) -> Collection[byotrack.Track]:
        try:
            if self.full_specs:
                self.full_specs.to_xml(detections_sequence).write(self.parameters_file)
            icy.save_detections(detections_sequence, self.rois_file)
            self._run_icy()
            # Sort tracks by starting time and then position (They seem to be randomly sorted otherwise and in addition
            # with EMC2 torch-tps approximate propagation it yields undeterministic behaviors)
            return sorted(
                icy.load_tracks(self.tracks_file), key=lambda track: (track.start, track.points[0].sum().item())
            )
        finally:
            if os.path.exists(self.rois_file):
                os.remove(self.rois_file)
            if os.path.exists(self.parameters_file):
                os.remove(self.parameters_file)
            if os.path.exists(self.tracks_file):
                os.remove(self.tracks_file)

    def _run_icy(self):
        """Run the icy process"""

        if self.full_specs:
            if self.full_specs.motion != self.motion:
                warnings.warn(
                    f"""motion attribute ({self.motion}) != full_specs.motion ({self.full_specs.motion})

Will use motion attribute (and modify the specifications)
                    """
                )

                self.full_specs.motion = self.motion

            self.runner.run(
                PARAMETRIZED_PROTOCOL,
                rois=os.path.abspath(self.rois_file),
                parameters=os.path.abspath(self.parameters_file),
                tracks=os.path.abspath(self.tracks_file),
            )
        else:
            directed, multi = self.motion.value

            self.runner.run(
                PROTOCOL,
                rois=os.path.abspath(self.rois_file),
                tracks=os.path.abspath(self.tracks_file),
                directed=directed,
                multi=multi,
            )

        # Has to check because icy do not return any non-zero return code
        if not os.path.exists(self.tracks_file):
            raise RuntimeError(
                """No track found from Icy.

            This probably results from a failure in Icy software. Please look at the Java Exception displayed by Icy.

            Note that Icy may fail the first time you launch this code because Icy updates the required protocols
            and is unable to continue to run after the update.
            """
            )



from __future__ import annotations

import dataclasses
import os
from typing import overload, Optional, Sequence, Tuple, Union

import numpy as np

from .reader import VideoReader
from .transforms import ChannelSelect, ChannelAvg, ScaleAndNormalize


@dataclasses.dataclass
class VideoTransformConfig:
    """Configuration for video transformations

    Attributes:
        aggregate (bool): Aggregate channels
        normalize (bool): Scale and Normalize the video in [0, 1]
        selected_channel (Optional[int]): Channel to use for aggregation
            If None, channel average is done. If any, it performs channel selection
        q_min (float): Minimum quantile to use when scaling the video
        q_max (float): Maximum quantile to use when scaling the video
        smooth_clip (float): Smoothness of the clipping process (scaling)

    """

    aggregate: bool = False
    normalize: bool = False
    selected_channel: Optional[int] = None
    q_min: float = 0.0
    q_max: float = 1.0
    smooth_clip: float = 0.0


class Video(Sequence[np.ndarray]):
    """Video: Iterable, indexable and sliceable sequence of frames wrapping a VideoReader

    It wraps VideoReader in order to add video transformation (Channel Aggregation, Scaling, Normalization)
    and to add useful pythonic protocols (Sliceable, Indexable, Iterable).

    Return images in BGR by default like opencv as frames are mostly used with opencv afterwards for display.
    Can also return grayscale image (H, W, 1)

    Example:
        .. code-block:: python

            import byotrack

            # Read a video (Usually BGR)
            video = byotrack.Video(video_path)

            # Add a transform that will aggregate channel and normalize in [0, 1] the intensities
            transform_config = byotrack.VideoTransformConfig(aggregate=True, normalize=True, q_min=0.01, q_max=0.999)
            video.set_transform(transform_config)

            # Iterate through the video
            for frame in video:
                pass

            # Temporal slicing

            sliced = video[10:50:3]  # Take one frame every three from frame 10 to frame 50.


    Attributes:
        shape (Tuple[int, int, int, int]): Shape of the video (Time, Height, Width, Channel)
        reader (byotrack.VideoReader): Underlying video reader

    """

    def __init__(self, data_source: Union[str, os.PathLike, VideoReader]) -> None:
        super().__init__()

        if isinstance(data_source, (str, os.PathLike)):
            self.reader = VideoReader.open(data_source)
        else:
            self.reader = data_source

        self._slice: Tuple[int, int, int] = (0, self.reader.length, 1)
        self._channel_aggregator: Optional[Union[ChannelAvg, ChannelSelect]] = None
        self._normalizer: Optional[ScaleAndNormalize] = None

    @property
    def shape(self) -> Tuple[int, int, int, int]:
        return (len(self), *self.reader.shape, self.reader.channels)

    def set_transform(self, transform_config: VideoTransformConfig) -> None:
        """Set the transform (channel_selector and normalizer)

        Args:
            config (VideoTransformConfig): Configuration of the transformations

        """
        self._channel_aggregator = None
        if transform_config.aggregate:
            if transform_config.selected_channel is not None:
                self._channel_aggregator = ChannelSelect(transform_config.selected_channel)
            else:
                self._channel_aggregator = ChannelAvg()

        self._normalizer = None
        if transform_config.normalize:
            self._normalizer = ScaleAndNormalize(
                transform_config.q_min, transform_config.q_max, transform_config.smooth_clip
            )
            self._set_normalizer()  # Set normalizer stats

    def _set_normalizer(self) -> None:
        """Set the normalizer stats (after aggregation)"""
        assert self._normalizer

        frame_id = self.reader.tell()
        self.reader.seek(self._slice[0])

        frames = []

        has_next = True
        while has_next:
            frame, has_next = self.reader.read()
            if self._channel_aggregator:
                frame = self._channel_aggregator(frame)

            frames.append(frame[None, ...])
            if len(frames) >= self._normalizer.max_frames_for_stats:
                break

        self._normalizer.update_stats(np.concatenate(frames, axis=0))

        self.reader.seek(frame_id)  # Reset reader where it was

    def transform(self, frame: np.ndarray) -> np.ndarray:
        """Transform a frame using channel aggregation and normalization"""
        if self._channel_aggregator:
            frame = self._channel_aggregator(frame)
        if self._normalizer:
            frame = self._normalizer(frame)

        return frame

    def __len__(self) -> int:
        return (self._slice[1] - self._slice[0] - 1) // self._slice[2] + 1  # Number of elt in the slice

    @overload
    def __getitem__(self, index: int) -> np.ndarray:
        ...

    @overload
    def __getitem__(self, slice_: slice) -> Video:
        ...

    def __getitem__(self, key):
        """Indexing and slicing operations

        When indexed, it returns the ith frame in the slice
        When sliced, it duplicates the video (wrapper) with the right slice

        Args:
            key (int | slice): index or slice of the video

        Returns:
            np.ndarray | Video: Frame at index or a shallow copy of the video with the right slice

        """
        if isinstance(key, int):
            frame_id = self._slice[0] + key * self._slice[2]

            if frame_id >= self._slice[1]:
                raise IndexError("Frame index out of range")

            if frame_id == self.reader.frame_id:
                pass  # Skip expensive seek
            if frame_id == self.reader.frame_id + 1:
                if not self.reader.grab():  # Much faster for cv2: Allows a fast frame by frame reading
                    raise IndexError("Frame index out of video")
            else:
                try:
                    self.reader.seek(frame_id)
                except EOFError as eof:
                    raise IndexError(eof.args) from None

            return self.transform(self.reader.retrieve())

        if isinstance(key, slice):
            start, stop, step = key.indices(len(self))  # Return int pos, int pos, int

            # Duplicate
            other = Video(self.reader)
            other._channel_aggregator = self._channel_aggregator
            other._normalizer = self._normalizer

            # Set the new slice
            other._slice = (
                self._slice[0] + start * self._slice[2],
                self._slice[0] + stop * self._slice[2],
                step * self._slice[2],
            )

            return other

        raise TypeError(f"Cannot index Video with type: {type(key)}")



from typing import Callable, Optional, Union
import warnings

import torch
import torch_tps
import tqdm


def forward_backward_propagation(
    tracks_matrix: torch.Tensor, method: Union[str, Callable[..., torch.Tensor]], **kwargs
) -> torch.Tensor:
    """Fill all NaN values in the tracks matrix by merging a forward and backward propagation

    Esimate missing positions for any defined track and frame. First it computes forward estimations,
    then backward ones. If estimations overlap (for instance with missing positions inside a track), it will
    merge smoothly the estimations.

    We provide two implementations for directed propagation:
    Constant where we propagate the last known position of the track
    TPS where we use ThinPlateSpline using the active tracks to estimate the motion of other tracks

    Args:
        tracks_matrix (torch.Tensor): Tracks data in a single tensor (See `Track.tensorize`)
            Shape: (T, N, D), dtype: float32
        method (str | Callable[..., torch.Tensor]): Method to use for propagation. Either "tps", "constant"
            of a self defined function to be called.
        **kwargs (Any): Additional arguments given to the method

    Returns:
        torch.Tensor: Extrapolated point for each track and time
            Shape: (T, N, D), dtype: float32
    """
    directed_propagate: Callable[..., torch.Tensor]
    if isinstance(method, str):
        if method.lower() == "tps":
            directed_propagate = tps_directed_propagate
        elif method.lower() == "constant":
            directed_propagate = constant_directed_propagate
        else:
            raise ValueError(f"Unknown method {method}. We only provide two of them: constant and tps")
    else:
        directed_propagate = method

    forward_positions = directed_propagate(tracks_matrix, True, **kwargs)
    backward_positions = directed_propagate(tracks_matrix, False, **kwargs)

    return merge(tracks_matrix, forward_positions, backward_positions)


def merge(
    tracks_matrix: torch.Tensor, forward_positions: torch.Tensor, backward_positions: torch.Tensor
) -> torch.Tensor:
    """Merge forward and backward propagation into a single propagation matrix

    If both propagation defines a position estimation, a weighted average is performed.

    Args:
        tracks_matrix (torch.Tensor): Original tracks data (`Track.tensorize`)
            Shape: (T, N, D), dtype: float32
        forward_positions (torch.Tensor): Forward estimation of positions
            Shape: (T, N, D), dtype: float32
        backward_positions (torch.Tensor): backward estimation of positions
            Shape: (T, N, D), dtype: float32

    Returns:
        torch.tensor: Merged propagation matrix
            Shape: (T, N, D), dtype: float32
    """

    # 1- Take forward and complete by backward

    propagation_matrix = forward_positions.clone()
    propagation_matrix[torch.isnan(forward_positions)] = backward_positions[torch.isnan(forward_positions)]

    # 2- If both forward and backward, merge them by a weighted sum
    both = (
        ~torch.isnan(backward_positions).any(dim=-1)
        & ~torch.isnan(forward_positions).any(dim=-1)
        & torch.isnan(tracks_matrix).any(dim=-1)
    )
    if both.any():
        undefined = torch.isnan(tracks_matrix).any(dim=-1)
        forward_score = torch.zeros(undefined.shape, dtype=torch.int16)[..., None]
        backward_score = torch.zeros(undefined.shape, dtype=torch.int16)[..., None]

        for i in range(1, tracks_matrix.shape[0]):
            # Forward (resp. backward) computation for backward score (resp. forward)
            backward_score[i, undefined[i]] = backward_score[i - 1, undefined[i]] + 1

            i = tracks_matrix.shape[0] - i - 1
            forward_score[i, undefined[i]] = forward_score[i + 1, undefined[i]] + 1

        propagation_matrix[both] = (
            forward_score[both] * forward_positions[both] + backward_score[both] * backward_positions[both]
        ) / (forward_score[both] + backward_score[both])

    return propagation_matrix


def constant_directed_propagate(tracks_matrix: torch.Tensor, forward=True) -> torch.Tensor:
    """Propagate tracks matrix with the last known position in a single direction

    Args:
        tracks_matrix (torch.Tensor): Tracks data in a single tensor (See `Tracks.tensorize`)
            Shape: (T, N, D), dtype: float32
        forward (bool): Forward or backward propagation
            Default: True (Forward)

    Returns:
        torch.Tensor: Estimation of tracks point in a single direction
            Shape: (T, N, D), dtype: float32
    """
    tracks_matrix = tracks_matrix if forward else torch.flip(tracks_matrix, (0,))

    propagation_matrix = tracks_matrix.clone()  # (T, N, D)
    valid = ~torch.isnan(propagation_matrix).any(dim=-1)  # (T, N)

    for i in tqdm.trange(1, tracks_matrix.shape[0], desc=f"{'Forward' if forward else 'Backward'} propagation"):
        # Compute propagation mask (not valid and has a past)
        propagation_mask = ~valid[i] & ~torch.isnan(propagation_matrix[i - 1]).any(dim=-1)

        if propagation_mask.sum() == 0:
            continue  # No propagation to do

        # Propagate points
        propagation_matrix[i, propagation_mask] = propagation_matrix[i - 1, propagation_mask]

    return propagation_matrix if forward else torch.flip(propagation_matrix, (0,))


def tps_directed_propagate(
    tracks_matrix: torch.Tensor, forward=True, alpha=5.0, track_mask: Optional[torch.Tensor] = None
) -> torch.Tensor:
    """Propagate tracks matrix using Thin Plate Spline (TPS) algorithm in a single direction

    .. note:: We use torch-tps which is very fast but not very accurate. For instance, shuflling the tracks
        may yield different results (<1px deviation). We have not noticed any real impact on EMC2 stitching
        performances. Alpha > 5.0 is advised (reduces the numerical errors + outliers resilience).


    Args:
        tracks_matrix (torch.Tensor): Tracks data in a single tensor (See `Tracks.tensorize`)
            Shape: (T, N, D), dtype: float32
        forward (bool): Forward or backward propagation
            Default: True (Forward)
        alpha (float): Regularization parameter of TPS
            Default: 5.0 (alpha > 5.0 is advised)
        track_mask (Optional[torch.Tensor]): Filter out some tracks to build control points
            Allow to drop uncertain tracks or to simulate propagation with less particles
            Shape: (N, ), dtype: bool

    Returns:
        torch.Tensor: Estimation of tracks point in a single direction
            Shape: (T, N, D), dtype: float32
    """
    tps = torch_tps.ThinPlateSpline(alpha, "cpu")  # Cf torch_tps: Faster on cpu than gpu

    tracks_matrix = tracks_matrix if forward else torch.flip(tracks_matrix, (0,))

    propagation_matrix = tracks_matrix.clone()  # (T, N, D)
    valid = ~torch.isnan(propagation_matrix).any(dim=-1)  # (T, N)

    for i in tqdm.trange(1, tracks_matrix.shape[0], desc=f"{'Forward' if forward else 'Backward'} propagation"):
        # Compute control mask (Valid before and now)
        control_mask = valid[i - 1] & valid[i]  # (N,)
        if track_mask is not None:
            control_mask &= ~track_mask  # Keep only unmasked tracks

        # Compute propagation mask (not valid and has a past)
        propagation_mask = ~valid[i] & ~torch.isnan(propagation_matrix[i - 1]).any(dim=-1)

        if propagation_mask.sum() == 0:
            continue  # No propagation to do

        if control_mask.sum() < 5:
            warnings.warn(f"Too few control points while propagating ({control_mask}.sum())")

        # Fit TPS on the control tracks from source and target frame
        tps.fit(tracks_matrix[i - 1, control_mask], tracks_matrix[i, control_mask])

        # Propagate points
        propagation_matrix[i, propagation_mask] = tps.transform(propagation_matrix[i - 1, propagation_mask])

    return propagation_matrix if forward else torch.flip(propagation_matrix, (0,))



"""Utilities for inputs/outputs with icy"""


import os
from typing import Collection, Union
from xml.etree import ElementTree as ET

import numpy as np
import tifffile  # type: ignore
import torch

import byotrack


def save_detections(detections_sequence: Collection[byotrack.Detections], path: Union[str, os.PathLike]) -> None:
    """Save a sequence of detections as a stack of segmentation that can be reload by the trackmate label detector

    .. warning:: Only supports consecutive detections

    Args:
        detections_sequence (Collection[Detections]): Detections for each frame (Should be consecutives)
        path (str | os.PathLike): Output path

    """
    segmentations = np.concatenate([detections.segmentation[None].numpy() for detections in detections_sequence])
    tifffile.imwrite(path, segmentations, photometric="minisblack")


def load_tracks(path: Union[str, os.PathLike]) -> Collection[byotrack.Track]:
    """Load tracks saved by trackmate

    Currently only 2d tracks are supported

    Format example:

    .. code-block:: xml

        <TrackMate>
            <Model>
                <AllSpots>
                    <SpotsInFrame frame="0">
                        <Spot ID="2811" POSITION_X="552.0" POSITION_Y="143.0" POSITION_Z="0.0">...</Spot>
                        ...
                    </SpotsInFrame>
                    ...
                </AllSpots>
                <AllTracks>
                    <Track TRACK_ID="0">
                        <Edge SPOT_SOURCE_ID="24068" SPOT_TARGET_ID="20796"/>
                        ...
                    </Track>
                </AllTracks>
            </Model>
        </TrackMate>

    Args:
        path (str | os.PathLike): Input path

    Returns:
        Collection[Track]: Parsed tracks

    """
    tracks = []
    model = ET.parse(path).getroot().find("Model")
    assert model is not None, "Data not found in file"

    # Load all spots to rebuild tracks
    spots = {}
    for frame_spots in model.findall("AllSpots")[0]:
        for spot in frame_spots:
            spots[spot.get("ID")] = (
                int(frame_spots.attrib["frame"]),
                float(spot.attrib["POSITION_X"]),
                float(spot.attrib["POSITION_Y"]),
            )

    # Go through tracks and rebuild from edges
    for track in model.findall("AllTracks")[0].findall("Track"):
        track_spots = set(map(lambda edge: spots[edge.get("SPOT_SOURCE_ID")], track.findall("Edge")))
        track_spots.update(map(lambda edge: spots[edge.get("SPOT_TARGET_ID")], track.findall("Edge")))
        spot_data = sorted(track_spots)
        start = spot_data[0][0]

        old_frame = start - 1
        points = []
        for frame, x, y in spot_data:
            # Extend with nan if there is a temporal gap
            assert frame > old_frame, "Splitting tracks are not supported (mutliple nodes at the same time)"
            points.extend([[torch.nan, torch.nan]] * (frame - old_frame - 1))
            points.append([y, x])
            old_frame = frame

        tracks.append(byotrack.Track(start, torch.tensor(points), int(track.attrib["TRACK_ID"])))

    return tracks


# Tracks could be saved in a simpler format by TrackMate with ExportTracksToXML.
# But this format is the one used in GUI, so let's keept it



import dataclasses
import json
import os
import pathlib
from typing import Collection, Iterable, Optional, Union

import numpy as np

import byotrack
from byotrack import fiji


SCRIPT = pathlib.Path(__file__).parent / "_trackmate.py"


@dataclasses.dataclass
class TrackMateParameters:  # pylint: disable=too-many-instance-attributes
    """Parameters of TrackMate [7, 8]

    We do *not* support Features penalties yet. Moreover track splitting and merging is currently *not* supported in
    ByoTrack.

    To use kalman filtering in addition to sparse lap tracking, set the `kalman_search_radius` value. By default,
    no kalman filtering is used.

    The main parameters to set are:

    * `linking_max_distance`: The distance threshold for frame-to-frame linking.
    * `max_frame_gap` and `gap_closing_max_distance`: The temporal and spatial distances for track stitching.
    * `kalman_search_radius`: The distance threshold for kalman linking (Switch to AdvancedKalmanTracker).

    Attributes:
        allow_gap_closing (bool): Use a second lap to solve tracklet stitching (Closing temporal gap
            between previously established tracks)
            Default: True
        allow_track_splitting (bool): NOT SUPPORTED (Useful for cell tracking)
            Default: False
        allow_track_merging (bool): NOT SUPPORTED (To allow split and remerge and handle split detections)
            Default: False
        alternative_linking_cost_factor (float): Alternative linking cost
            Default: 1.05
        blocking_value (float): Cost for mon-physical, forbidden links.
            Default: inf
        cutoff_percentile (float): Cutoff percentile
            Default: 0.9
        max_frame_gap (int): The max difference in frames between two spots to allow for linking.
            For instance a value of 2 means that the tracker will be able to make a link between a spot in frame t
            and a successor spots in frame t+2, effectively bridging over one missed detection in one frame.
            Default: 2 (1 missed detections)
        linking_max_distance (float): The max distance between two consecutive spots, in physical units, allowed
            for creating links. If using kalman filters: this is the initial search radius, in physical units,
            specifying how far two spots can be apart when initiating new tracks. (See `kalman_search_raius`)
            Default: 15.0
        gap_closing_max_distance (float): Gap-closing max spatial distance. The max distance between two spots,
            in physical units, allowed for creating links over missing detections.
            Default: 15.0
        merging_max_distance (float): Unused. Track merging max spatial distance.
            Default: 15.0
        splitting_max_distance (float): Unused. Track splitting max spatial distance.
            Default: 15.0
        kalman_search_radius (Optional[float]): Set this parameter to use the AdvancedKalmanTracker. The max search
            radius specifying how far from a predicted position the tracker should look for candidate spots.
            Default: None (Without kalman filters)

    """

    allow_gap_closing: bool = True
    # Track merging/splitting is not supported
    allow_track_splitting: bool = False
    allow_track_merging: bool = False

    alternative_linking_cost_factor: float = 1.05
    blocking_value: float = float("inf")
    cutoff_percentile: float = 0.9

    max_frame_gap: int = 2

    linking_max_distance: float = 15.0
    gap_closing_max_distance: float = 15.0
    merging_max_distance: float = 15.0
    splitting_max_distance: float = 15.0

    # linking_feature_penalties
    # gap_closing_feature_penalties
    # merging_feature_penalties
    # splitting_feature_penalties

    kalman_search_radius: Optional[float] = None

    def save(self, path: Union[str, os.PathLike]) -> None:
        """Save the parameters to the given path

        Uses json format.

        Args:
            path (str | os.PathLike): Path to the output file
        """
        with open(path, "w", encoding="utf-8") as file:
            json.dump(dataclasses.asdict(self), file)


class TrackMateLinker(byotrack.Linker):  # pylint: disable=too-few-public-methods
    """Run TrackMate [7, 8] from Fiji [6]

    Wrapper around TrackMate, using Fiji headless call/

    About TrackMate:
    It is a global distance minimization tracking. It supports multiple algorithms. We have wrapped the more advanced
    ones (SparseLapTracker and AdvancedKalmanTracker that both follows [7]). It solves frame-to-frame GDM between
    detections. And then sovle a GDM between tracklets to correct them (stitch, merge, split).
    The AdvancedKalmanTracker additionnaly uses kalman filters to estimate velocities of particles.

    Here we rely on the handmade ImageJ script "_trackmate.py" that expects detections as instance segmentation
    and the hyperparameters of the linking process.

    We do not support track splitting and merging yet, neither the use of feature-based cost.

    The workflow is:

    1. Save detections to segmentation format
    2. Run the Fiji script
    3. [In Fiji] Read segmentation, load parameters, extract detections, run linking [5], export tracks to xml
    4. Read Fiji tracks and return

    Note:
        This implementation requires Fiji to be installed (https://imagej.net/downloads)

    Note:
        In case of missed detections, positions are filled with nan. To fill nan with true values, use an Interpolator

    Attributes:
        runner (byotrack.fiji.FijiRunner): Fiji runner
        specs (TrackmateParameters): Parameters specifications of the algorithm

    """

    detections_file = "_tmp_detections.tif"
    parameters_file = "_tmp_parameters.json"
    tracks_file = "_tmp_tracks.xml"

    def __init__(self, fiji_path: Union[str, os.PathLike], specs: TrackMateParameters) -> None:
        """Constructor

        Args:
            fiji_path (str | os.PathLike): Path to the fiji executable
                The executable can be found inside the installation folder of Fiji.
                Linux: Fiji.app/ImageJ-<os>.exe
                Windows: Fiji.app/ImageJ-<os>.exe
                MacOs: Fiji.app/Contents/MacOs/ImageJ-<os>

        """
        super().__init__()
        self.runner = fiji.FijiRunner(fiji_path)
        self.specs = specs

    def run(
        self, video: Iterable[np.ndarray], detections_sequence: Collection[byotrack.Detections]
    ) -> Collection[byotrack.Track]:
        try:
            self.specs.save(self.parameters_file)
            fiji.save_detections(detections_sequence, self.detections_file)
            self._run_fiji()
            # Sort tracks by starting time and then position (Prevents undeterministic behaviors with some post
            # processing steps)
            return sorted(
                fiji.load_tracks(self.tracks_file), key=lambda track: (track.start, track.points[0].sum().item())
            )
        finally:
            if os.path.exists(self.detections_file):
                os.remove(self.detections_file)
            if os.path.exists(self.parameters_file):
                os.remove(self.parameters_file)
            if os.path.exists(self.tracks_file):
                os.remove(self.tracks_file)

    def _run_fiji(self):
        """Run the fiji process"""
        self.runner.run(
            SCRIPT,
            detections=os.path.abspath(self.detections_file),
            parameters=os.path.abspath(self.parameters_file),
            tracks=os.path.abspath(self.tracks_file),
        )

        # Has to check because fiji do not return any non-zero return code
        if not os.path.exists(self.tracks_file):
            raise RuntimeError(
                """No track found from Fiji.

            This probably results from a failure inside Fiji script.
            Please look at the Java Exception displayed by Fiji.
            """
            )



from typing import Callable, Collection, Iterable, List, Union

import numpy as np
import torch

import byotrack

from . import propagation


class ForwardBackwardInterpolater(byotrack.Refiner):  # pylint: disable=too-few-public-methods
    """Interpolate tracks to fill out NaN values by averaging a forward estimation and a backward estimation.

    We propose two modes for position propagation: constant and tps. Constant estimation propagates the
    last known position, whereas tps uses active tracks knowledge to estimate the elastic motion.

    Forward and backward propagation are smoothly merged together in a weighted average. See `propagation` module

    Attributes:
        method (str | Callable[..., torch.Tensor]): Method for directed propagation
            ("constant", "tps" or your own callable)
            Default: "constant"
        full (bool): Interpolate track on all frames or only on the track range
            Default: False (Will not extend partial tracks, but only fill out NaN values inside of them)
        **kwargs: Additional arguments given to the method (For instance, `alpha` for tps)

    """

    def __init__(self, method: Union[str, Callable[..., torch.Tensor]] = "constant", full=False, **kwargs) -> None:
        super().__init__()
        self.method = method
        self.full = full
        self.kwargs = kwargs

    def run(self, video: Iterable[np.ndarray], tracks: Collection[byotrack.Track]) -> List[byotrack.Track]:
        if not tracks:
            return []

        tracks_matrix = byotrack.Track.tensorize(tracks)
        propagation_matrix = propagation.forward_backward_propagation(tracks_matrix, self.method, **self.kwargs)

        new_tracks = []
        for i, track in enumerate(tracks):
            if self.full:
                start = 0
                points = propagation_matrix[:, i]
            else:
                start = track.start
                points = propagation_matrix[start : start + len(track), i]
            new_tracks.append(byotrack.Track(start, points, track.identifier))

        return new_tracks



# @String detections
# @String parameters
# @String tracks

# pylint: skip-file
# type: ignore

import json
import sys

from java.io import File

from ij import IJ

from fiji.plugin.trackmate import Model
from fiji.plugin.trackmate import Settings
from fiji.plugin.trackmate import TrackMate
from fiji.plugin.trackmate import Logger
from fiji.plugin.trackmate.io import TmXmlWriter
from fiji.plugin.trackmate.action import ExportTracksToXML
from fiji.plugin.trackmate.detection import LabelImageDetectorFactory
from fiji.plugin.trackmate.tracking.jaqaman import SparseLAPTrackerFactory
from fiji.plugin.trackmate.tracking.kalman import AdvancedKalmanTrackerFactory

# We have to do the following to avoid errors with UTF8 chars generated in
# TrackMate that will mess with our Fiji Jython.
reload(sys)
sys.setdefaultencoding("utf-8")


print("Hello from ImageJ/Fiji")

print("Loading detections from", detections)

# Load detections as an image (As we use LabelImageDetector)
imp = IJ.openImage(detections)
imp.setStack(imp.getStack(), 1, 1, imp.getStackSize())
imp.show()

# ----------------------------
# Create the model object now
# ----------------------------

# Some of the parameters we configure below need to have
# a reference to the model at creation. So we create an
# empty model now.

model = Model()

# Send all messages to ImageJ log window.
model.setLogger(Logger.DEFAULT_LOGGER)


# ------------------------
# Prepare settings object
# ------------------------

settings = Settings(imp)

# Configure detector - Let's use a LabelImageDetector (detections is already done)
settings.detectorFactory = LabelImageDetectorFactory()
settings.detectorSettings = {
    "TARGET_CHANNEL": 1,
    "SIMPLIFY_CONTOURS": False,
}

# Configure tracker from the parameters

## First: Load the parameters
with open(parameters, "r") as file:
    specs = json.load(file)

if specs["kalman_search_radius"] is None:  # Without kalman -> Use SparseLAPTracker
    settings.trackerFactory = SparseLAPTrackerFactory()
    specs.pop("kalman_search_radius")
else:
    settings.trackerFactory = AdvancedKalmanTrackerFactory()

## Generate default settings
settings.trackerSettings = settings.trackerFactory.getDefaultSettings()

# Override settings
for key, value in specs.items():
    settings.trackerSettings[key.upper()] = value

print("Settings:")
print(settings.trackerSettings)


# -------------------
# Instantiate plugin
# -------------------

trackmate = TrackMate(model, settings)

# --------
# Process
# --------

ok = trackmate.checkInput()
if not ok:
    sys.exit(str(trackmate.getErrorMessage()))

ok = trackmate.process()
if not ok:
    sys.exit(str(trackmate.getErrorMessage()))


# -------------
# Save results
# -------------

# Could use this function as it outputs a cleaner results... but not the one used in GUI so lets keep the other one
# ExportTracksToXML.export(model, settings, File(tracks))

writer = TmXmlWriter(File(tracks))
writer.appendSettings(settings)
writer.appendModel(model)
writer.writeToFile()



from typing import Collection, Optional, Sequence, Tuple

import cv2
import matplotlib as mpl  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import numpy as np
import torch

import byotrack


hsv = mpl.colormaps["hsv"]
colors = list(map(lambda x: tuple(int(c * 255) for c in x[:3]), map(hsv, [i / 200 for i in range(200)])))


def display_lifetime(tracks: Collection[byotrack.Track]):
    """Display the lifetime of tracks

    Args:
        tracks (Collection[byotrack.Track]): Tracks

    """
    tracks_tensor = byotrack.Track.tensorize(tracks)

    mask = (~torch.isnan(tracks_tensor).any(dim=2, keepdim=True)).numpy() * 255
    mask = np.concatenate([mask] * 3, axis=2)

    plt.figure(figsize=(24, 16), dpi=100)
    plt.xlabel("Track id")
    plt.ylabel("Frame")
    plt.imshow(mask)
    plt.show()


class InteractiveVisualizer:
    """Interactive visualization with opencv

    Keys:
        * `space`: Pause/Unpause the video
        * w/x: Move backward/forward in the video (when paused)
        * d: Switch detections display mode (Not displayed, Mask, Segmentation) if available
        * t: Switch on/off the display of tracks if available
        * v: Switch on/off the display of the video

    Attributes:
        video (Optional[Sequence[np.ndarray]]): Optional video to display. Should be normalized in [0, 1]
            Default: None
        detections_sequence (Collection[byotrack.Detections]): Optional detections to display
            Default: () (no detections)
        tracks (Collection[byotrack.Tracks]): Optional tracks to display
            Default: () (no tracks)
        frame_shape (Tuple[int, int]): Shape of frames
        n_frames (int): Number of frames
    """

    window_name = "ByoTrack ViZ"
    tracks_colors = colors

    def __init__(
        self,
        video: Optional[Sequence[np.ndarray]] = None,
        detections_sequence: Collection[byotrack.Detections] = (),
        tracks: Collection[byotrack.Track] = (),
    ) -> None:
        assert video is not None or detections_sequence or tracks, "No data to display"

        self.video = video
        self.detections_sequence = detections_sequence
        self.tracks = tracks

        self.frame_shape = self._get_frame_shape(self.video, self.detections_sequence, self.tracks)
        self.n_frames = self._get_n_frames(self.video, self.detections_sequence, self.tracks)

        self._frame_id = 0
        self._display_video = int(video is not None)
        self._display_detections = int(bool(detections_sequence))
        self._display_tracks = int(bool(tracks))
        self._running = False

    def run(self, frame_id=0, fps=20) -> None:
        """Run the visualization

        Args:
            frame_id (int): Starting frame_id
            fps (int): Frame rate
        """
        try:
            self._frame_id = frame_id
            self._run(fps)
        finally:
            cv2.destroyWindow(self.window_name)

    def _run(self, fps=20) -> None:
        frame_to_detections = {detections.frame_id: detections for detections in self.detections_sequence}

        while True:
            self._frame_id += self._running
            frame = np.zeros((*self.frame_shape, 3), dtype=np.uint8)

            if self._display_video and self.video is not None:
                frame = (self.video[self._frame_id] * 255).astype(np.uint8)
                assert frame.shape[2] in (1, 3)

                if frame.shape[2] == 1:
                    frame = np.concatenate([frame] * 3, axis=2)

            if self._display_detections and self._frame_id in frame_to_detections:
                segmentation: np.ndarray = frame_to_detections[self._frame_id].segmentation.clone().numpy()
                if self._display_detections == 1:  # Display segmentation as mask
                    segmentation = segmentation != 0
                    segmentation = segmentation.astype(np.uint8)[..., None] * 255
                else:
                    segmentation = (segmentation % 206) + 50
                    segmentation[frame_to_detections[self._frame_id].segmentation == 0] = 0
                    segmentation = segmentation.astype(np.uint8)[..., None]

                segmentation = np.concatenate(
                    [segmentation, np.zeros_like(segmentation), np.zeros_like(segmentation)], axis=2
                )

                if self._display_video:
                    frame = frame / 2 + segmentation / 2
                    frame = frame.astype(np.uint8)
                else:
                    frame = segmentation

            if self._display_tracks:
                for track in self.tracks:
                    point = track[self._frame_id]
                    if torch.isnan(point).any():
                        continue

                    i, j = point.round().to(torch.int).tolist()

                    color = colors[track.identifier % len(colors)]

                    cv2.circle(frame, (j, i), 5, color)
                    cv2.putText(
                        frame, str(track.identifier % 100), (j + 4, i - 4), cv2.FONT_HERSHEY_SIMPLEX, 0.3, color
                    )

            # Display the resulting frame
            cv2.imshow(self.window_name, frame)
            cv2.setWindowTitle(self.window_name, f"Frame {self._frame_id} / {self.n_frames}")

            # Handle user actions
            key = cv2.waitKey(1000 // fps) & 0xFF
            if self.handle_actions(key):
                break

    def handle_actions(self, key: int) -> bool:
        """Handle inputs from user

        Return True to quit

        Args:
            key (int): Key input from user

        Returns:
            bool: True to quit visualization
        """
        if key == ord("q"):
            return True

        if cv2.getWindowProperty(self.window_name, cv2.WND_PROP_VISIBLE) < 1:
            return True

        if key == ord(" "):
            self._running = not self._running

        if not self._running and key == ord("w"):  # Prev
            self._frame_id = (self._frame_id - 1) % self.n_frames

        if not self._running and key == ord("x"):  # Next
            self._frame_id = (self._frame_id + 1) % self.n_frames

        if key == ord("d"):
            self._display_detections = (self._display_detections + 1) % 3

        if key == ord("t"):
            self._display_tracks = 1 - self._display_tracks

        if key == ord("v"):
            self._display_video = 1 - self._display_video

        return False

    @staticmethod
    def _get_frame_shape(
        video: Optional[Sequence[np.ndarray]] = None,
        detections_sequence: Collection[byotrack.Detections] = (),
        tracks: Collection[byotrack.Track] = (),
    ) -> Tuple[int, int]:
        frame_shape: Tuple[int, int] = (1, 1)
        if video is not None:
            frame_shape = np.broadcast_shapes(frame_shape, video[0].shape[:2])  # type: ignore
        if detections_sequence:
            detections = next(iter(detections_sequence))
            frame_shape = np.broadcast_shapes(frame_shape, detections.shape)  # type: ignore
        if tracks:
            track_tensor = byotrack.Track.tensorize(tracks)
            positions = track_tensor[~torch.isnan(track_tensor).any(dim=2)]
            minimum: torch.Tensor = positions.min(dim=0).values.round()
            maximum: torch.Tensor = positions.max(dim=0).values.round()

            if frame_shape == (1, 1):
                size = (maximum + 20).to(torch.int).tolist()  # Add some margin
                frame_shape = (size[0], size[1])

            if (minimum < 0).any() or maximum[0] >= frame_shape[0] or maximum[1] >= frame_shape[1]:
                print("Some tracks are out of frame")

        return frame_shape

    @staticmethod
    def _get_n_frames(
        video: Optional[Sequence[np.ndarray]] = None,
        detections_sequence: Collection[byotrack.Detections] = (),
        tracks: Collection[byotrack.Track] = (),
    ) -> int:
        if video is not None:
            return len(video)

        n_frames = 1
        if tracks:
            n_frames = max(track.start + len(track) for track in tracks)

        if detections_sequence:
            n_frames = max(
                n_frames,
                1 + max(detections.frame_id if detections.frame_id else 0 for detections in detections_sequence),
            )

        return n_frames



from typing import Collection, Iterable, List

import numpy as np
import torch

import byotrack


class Cleaner(byotrack.Refiner):
    """Cleaner refiner

    Split tracklet when the distance between two consecutive points is greater than `max_dist`.
    And delete tracks that are shorter than `min_length`.

    Warning: Does not really support tracks with nan points (will just not split when a nan is met)

    Attributes:
        min_length (int): Minimum length of tracks kept
        max_dist (float): Maximum distance between two consecutive points in a track

    """

    def __init__(self, min_length: int, max_dist: float) -> None:
        super().__init__()
        self.min_length = min_length
        self.max_dist = max_dist

    def run(self, video: Iterable[np.ndarray], tracks: Collection[byotrack.Track]) -> List[byotrack.Track]:
        return self.clean_tracks(tracks, self.min_length, self.max_dist)

    @staticmethod
    def clean_tracks(tracks: Collection[byotrack.Track], min_length: int, max_dist: float) -> List[byotrack.Track]:
        """Clean tracks

        Split tracklet when the distance between two consecutive points is greater than `max_dist`.
        And delete tracks that are shorter than `min_length`.

        Warning: Does not really support tracks with nan points (will just not split when a nan is met)

        Args:
            tracks (Collection[Track]): Tracks to clean
            min_length (int): Minimum length of tracks kept
            max_dist (float): Maximum distance between two consecutive points in a track
                If max_dist <= 0, splitting is not done

        Returns:
            List[Tracklet]: Cleaned tracks

        """
        n_split = 0
        n_filtered = 0
        cleaned_tracks: List[byotrack.Track] = []

        for track in tracks:
            cleaned = []
            if max_dist <= 0:
                cleaned.append(track)
            else:
                speed = torch.norm(track.points[1:] - track.points[:-1], dim=1)
                first = 0
                for i in range(len(track) - 1):
                    if speed[i] > max_dist:  # Break (i -> i + 1)
                        cleaned.append(
                            byotrack.Track(
                                track.start + first,
                                track.points[first : i + 1],
                                identifier=track.identifier,
                            )
                        )
                        first = i + 1
                        n_split += 1

                # There is at least one element left
                cleaned.append(byotrack.Track(track.start + first, track.points[first:], identifier=track.identifier))

            n_filtered += len(cleaned)
            cleaned = list(filter(lambda tracklet: len(tracklet) >= min_length, cleaned))
            n_filtered -= len(cleaned)

            cleaned_tracks.extend(cleaned)

        print(f"Cleaning: Split {n_split} tracks and dropped {n_filtered} resulting ones")
        print(f"Cleaning: From {len(tracks)} to {len(cleaned_tracks)} tracks")
        return cleaned_tracks
